// Copyright (c) 2022 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "assert/assertgl.hpp"
#include "Shader.hpp"

using namespace Dawn;

void Shader::compileShader(
  std::map<std::string, int32_t> attributeLocations,
  const std::string vertexShader,
  const std::string fragmentShader
) {
  GLint isSuccess;
  int32_t maxLength;
  char error[1024];
  
  // Load the vertex shader first
  this->shaderVertex = glCreateShader(GL_VERTEX_SHADER);
  auto vertShaderC = vertexShader.c_str();
  glShaderSource(this->shaderVertex, 1, &vertShaderC, 0);
  assertNoGLError();
  glCompileShader(this->shaderVertex);
  assertNoGLError();

  // Validate
  glGetShaderiv(this->shaderVertex, GL_COMPILE_STATUS, &isSuccess);
  if(!isSuccess) {
    glGetShaderiv(this->shaderVertex, GL_INFO_LOG_LENGTH, &maxLength);
    glGetShaderInfoLog(this->shaderVertex, maxLength, &maxLength, error);
    assertUnreachable("Error compiling vert shader %s", error);
    throw error;
  }
  assertNoGLError();

  // Now load the Frag shader
  this->shaderFrag = glCreateShader(GL_FRAGMENT_SHADER);
  auto fragShaderC = fragmentShader.c_str();
  glShaderSource(this->shaderFrag, 1, &fragShaderC, 0);
  glCompileShader(this->shaderFrag);
  glGetShaderiv(this->shaderFrag, GL_COMPILE_STATUS, &isSuccess);
  if(!isSuccess) {
    glGetShaderiv(this->shaderFrag, GL_INFO_LOG_LENGTH, &maxLength);
    glGetShaderInfoLog(this->shaderFrag, maxLength, &maxLength, error);
    glDeleteShader(this->shaderVertex);
    assertUnreachable("Error compiling frag shader %s", error);
    throw error;
  }
  assertNoGLError();

  // Now create the shader program.
  shaderProgram = glCreateProgram();
  glAttachShader(shaderProgram, shaderVertex);
  glAttachShader(shaderProgram, shaderFrag);
  assertNoGLError();

  // Now parse out the variables.
  #if DAWN_OPENGL_HLSL
    auto itAttr = attributeLocations.begin();
    while(itAttr != attributeLocations.end()) {
      this->bindAttributeLocation(itAttr->first, itAttr->second);
      ++itAttr;
    }
  #endif

  //Bind, Verify & Use the shader program
  glLinkProgram(this->shaderProgram);
  glGetProgramiv(this->shaderProgram, GL_LINK_STATUS, &isSuccess);
  if(!isSuccess) {
    glGetProgramiv(this->shaderProgram, GL_INFO_LOG_LENGTH, &maxLength);
    glGetProgramInfoLog(this->shaderProgram, maxLength, &maxLength, error);
    glDeleteShader(this->shaderVertex);
    glDeleteShader(this->shaderFrag);
    assertUnreachable("Error compiling shader program %s", error);
    throw error;
  }
  assertNoGLError();
}

void Shader::bindAttributeLocation(
  const std::string name,
  const int32_t location
) {
  if(this->shaderProgram == -1) throw "Shader has not yet been compiled";
  glBindAttribLocation(this->shaderProgram, location, name.c_str());
  assertNoGLError();
}

void Shader::setTexture(
  const shaderparameter_t param,
  const textureslot_t slot
) {
  glUniform1i(param, slot);
  assertNoGLError();
}

shaderparameter_t Shader::getParameterByName(const std::string name) {
  return glGetUniformLocation(this->shaderProgram, name.c_str());
}

shaderbufferlocation_t Shader::getBufferLocationByName(const std::string name) {
  return glGetUniformBlockIndex(this->shaderProgram, name.c_str());
}

void Shader::setParameterBuffer(
  const shaderbufferlocation_t location,
  const shaderbufferslot_t slot
) {
  glUniformBlockBinding(this->shaderProgram, location, slot);
  assertNoGLError();
}

void Shader::setMatrix(
  const shaderparameter_t uniform,
  const glm::mat4 matrix
) {
  glUniformMatrix4fv(uniform, 1, GL_FALSE, glm::value_ptr(matrix));
  assertNoGLError();
}

void Shader::setBoolean(
  const shaderparameter_t uni, 
  const bool value
) {
  glUniform1i(uni, value);
  assertNoGLError();
}

void Shader::setColor(
  const shaderparameter_t uniform,
  const struct Color color
) {
  glUniform4f(uniform, color.r, color.g, color.b, color.a);
  assertNoGLError();
}

void Shader::setVector3(
  const shaderparameter_t uniform,
  const glm::vec3 vector
) {
  glUniform3f(uniform, vector.x, vector.y, vector.z);
  assertNoGLError();
}

void Shader::setFloat(
  const shaderparameter_t param,
  const float_t value
) {
  glUniform1f(param, value);
  assertNoGLError();
}

void Shader::bind() {
  assertTrue(shaderProgram != -1, "Cannot bind a program that is not ready");
  glUseProgram(shaderProgram);
  assertNoGLError();
}

Shader::~Shader() {
  if(shaderProgram != -1) glDeleteProgram(shaderProgram);
  if(shaderVertex != -1) glDeleteShader(shaderVertex);
  if(shaderFrag != -1) glDeleteShader(shaderFrag);
}